import asyncio
import json
import os
import re
from glob import glob

import backtrace_unreald.unreald_multipart_post as mp
import backtrace_unreald.unreald_utils as uu
from backtrace_unreald.unreald_utils import log


class UploadRequest:
    def __init__(self, name, dir):
        self.name = name
        self.dir = dir
        self.is_removed = False
        self.req = None

    def get_request_filename(self) -> str:
        """
        Gets filename of the request file.
        """

        self.__throw_if_removed()

        return '%s.req' % self.name

    def get_attachment_file_names(self) -> "list[str]":
        """
        Returns all attachments file names associated with this request.
        """

        self.__throw_if_removed()

        req = self.get_content()
        return list(req['files'])

    def get_file_names(self):
        """
        Returns all file names associated with this request.
        """

        self.__throw_if_removed()

        return self.get_attachment_file_names() + [self.get_request_filename()]

    def get_attachment_file_paths(self):
        """
        Gets relative paths to all attachments.
        """

        return [self.get_file_path(f) for f in self.get_attachment_file_names()]

    def get_content(self):
        """
        Loads the request JSON file and returns the object.
        """

        self.__throw_if_removed()

        if self.req is not None:
            return self.req

        path = self.get_file_path(self.get_request_filename())
        with open(path, 'r') as file:
            self.req = json.load(file)
        return self.req

    def get_file_path(self, file, dir = None):
        """
        Returns relative path to the file.
        """

        self.__throw_if_removed()

        return '%s/%s' % (self.dir if dir is None else dir, file)

    def move_to(self, destination):
        """
        Moves all request files to `destination`.
        """

        self.__throw_if_removed()

        if self.dir == destination:
            return

        for filename in self.get_file_names():
            old_path = self.get_file_path(filename)
            new_path = self.get_file_path(filename, destination)
            os.rename(old_path, new_path)
        self.dir = destination

    def remove(self):
        """
        Removes all files associated with this request.
        This request cannot be used after this operation.
        """

        self.__throw_if_removed()

        for filename in self.get_file_names():
            path = self.get_file_path(filename)
            os.remove(path)
        self.is_removed = True

    def __throw_if_removed(self):
        if self.is_removed:
            raise RuntimeError('Request "%s" is removed and cannot be used anymore.' % self.name)

class Uploader:
    def __init__(self, config, pending_dir, sending_dir):
        self.config = config
        self.semaphore = asyncio.Semaphore(self.config.limits.max_concurrent_uploads)
        self.pending_dir = pending_dir
        self.sending_dir = sending_dir

    def start(self):
        # Enqueue all not sent requests in sending_dir
        for num in self.get_requests(self.sending_dir):
            self.enqueue(num, self.sending_dir)

    def get_requests(self, dir):
        files = [os.path.basename(x) for x in glob('%s/*.req' % dir)]
        files = [int(m.group(1)) for f in files for m in [re.search('(\\d+).req', f)]]
        return files

    def enqueue(self, num, dir=None):
        asyncio.ensure_future(
            self.send_file(num, dir, retries=self.config.retries.count,
                                     retry_delay=self.config.retries.delay,
                                     retry_exp=self.config.retries.exp))

    def get_queue_length(self):
        return len(self.get_requests(self.pending_dir))

    @uu.report_to_backtrace_async()
    async def send_file(self, file_name_base, dir = None, retries = 0, retry_delay = 1, retry_exp = 1):
        request = UploadRequest(file_name_base, dir if dir is not None else self.pending_dir)
        while retries >= 0:
            try:
                async with self.semaphore:
                    log('Processing %s, available retries %d, current delay %ds' % (file_name_base, retries, retry_delay))
                    request.move_to(self.sending_dir)
                    await self.send_file_inner(request)
                    request.remove()
                    log('OnUploadSuccess %s' % file_name_base)
                    return
            except Exception as e:
                log('OnUploadFail %s %s' % (file_name_base, e))
                if retries == 0:
                    log('Processing %s failed. Maximum retries reached' % file_name_base)
                    request.remove()
                    raise

                log('Processing %s failed, retrying in %ds' % (file_name_base, retry_delay))
                await asyncio.sleep(retry_delay)
                retry_delay = retry_delay * retry_exp
                retries -= 1

    async def send_file_inner(self, request: UploadRequest):
        req = request.get_content()
        url = req['url']

        if self.is_blacklisted(url):
            log('Skipping request %s, because URL is blacklisted (url: %s)' % (request.name, url))
            return

        attributes = req['attributes']
        files = []
        for attachment_name in request.get_attachment_file_names():
            path = request.get_file_path(attachment_name)
            real_name = req['files'][attachment_name]

            if real_name.endswith('.dmp'):
                name = 'upload_file_minidump'
                files.insert(0, [name, name, path])
                continue

            files.append(['attachment_%s' % real_name, real_name, path])

        log('Sending crash report %s to %s' % (request.name, url))

        # await asyncio.get_event_loop().run_in_executor(None, mp.post, url, attributes, files)
        await mp.post_async(url, attributes, files)

    def is_blacklisted(self, url):
        for el in self.config.blacklist:
            if re.search(el.regex, url):
                log('Url matches blacklisted /%s/ (url: %s)' % (el.regex, url))
                return True
        return False
