import datetime
import grp
import logging
import os
import pwd
import sys

import backtracepython as bt
import tornado
import tornado.log

LOGGER = None

# Turn reporting on or off
REPORTING = True

def report_to_backtrace(func):
    def inner(*args, **kvargs):
        try:
            return func(*args, **kvargs)
        except Exception as e:
            if REPORTING:
                log('Error. Reporting %s' % e)
                bt.send_last_exception()
            else:
                log('Error: %s' % e)
            return None
    return inner

def report_to_backtrace_async():
    def wrapper(func):
        async def inner(*args, **kvargs):
            try:
                return await func(*args, **kvargs)
            except Exception as e:
                if REPORTING:
                    log('Error. Reporting %s' % e)
                    bt.send_last_exception()
                else:
                    log('Error: %s' % e)
                return None
        return inner
    return wrapper


def timestamp_ms():
    diff = datetime.datetime.utcnow() - datetime.datetime(1970, 1, 1)
    return int(diff.total_seconds() * 1000)

def timestamp_str():
    now = datetime.datetime.utcnow()
    return now.strftime('%Y-%m-%d %H:%M:%S.%f')

def init_logger():
    global LOGGER
    LOGGER = logging.getLogger()
    streaming_handler = logging.StreamHandler(stream=sys.stderr)
    LOGGER.addHandler(streaming_handler)
    LOGGER.setLevel(logging.INFO)
    tornado.log.access_log.addHandler(streaming_handler)


def log(msg):
    # msg = '[%s] %s' % (timestamp_str(), msg)
    # print(msg)
    LOGGER.error(msg)


def drop_privileges(uid_name='nobody', gid_name='nobody'):
    if os.getuid() != 0:
        # We're not root so, like, whatever dude
        return

    # Get the uid/gid from the name
    running_uid = pwd.getpwnam(uid_name).pw_uid
    running_gid = grp.getgrnam(gid_name).gr_gid

    # Remove group privileges
    os.setgroups([])

    # Try setting the new uid/gid
    os.setgid(running_gid)
    os.setuid(running_uid)

    # Ensure a very conservative umask
    old_umask = os.umask(0o77)
    return old_umask

def chown_dir(dir, uid_name='nobody', gid_name='nobody'):
    uid = pwd.getpwnam(uid_name).pw_uid
    gid = grp.getgrnam(gid_name).gr_gid

    log('chowning dir \'{dir}\' to {uid_name}({uid}), '\
        '{gid_name}({gid}).'.format(dir=dir,
                                    uid_name=uid_name,
                                    uid=uid,
                                    gid_name=gid_name,
                                    gid=gid))
    os.chown(dir, uid, gid)

# Calling it directly instantiates the object/logger and allows us to just import the log
# method only into each module.
# Note we can only call this after init_logger has been defined. Calling it up top won't work.
init_logger()
