import asyncio
import json
import os
import re
import zlib
from glob import glob

import tornado.httpclient
import tornado.template
import tornado.web

import backtrace_unreald.unreald_parse_xml as uxml
import backtrace_unreald.unreald_utils as uu
from backtrace_unreald.unreald_uploader import Uploader
from backtrace_unreald.unreald_utils import log

SIZEOF_INT = 4


@uu.report_to_backtrace
def ReadStringFromData(data, offset):
    header_size = 0
    str_length = int.from_bytes(
        data[offset:offset + SIZEOF_INT], byteorder='little')
    if str_length is None or str_length != 260:
        header_size = 3
        str_length = int.from_bytes(
            data[offset + 3:offset + SIZEOF_INT + 3], byteorder='little')
    # hack
    if str_length != 260:
        return None, None
    str_end_idx = offset + header_size + str_length + SIZEOF_INT
    str_data = data[offset + header_size +
                    SIZEOF_INT:str_end_idx].decode('utf-8').rstrip('\0')
    return str_data, str_end_idx


@uu.report_to_backtrace
def GetFileFromData(data, offset):
    file_length = int.from_bytes(
        data[offset:offset + SIZEOF_INT], byteorder='little')
    file_end_idx = offset + file_length + SIZEOF_INT
    file_data = data[offset + SIZEOF_INT:file_end_idx]
    return file_data, file_end_idx


@uu.report_to_backtrace
def SkipFileInData(data, offset):
    file_length = int.from_bytes(
        data[offset:offset + SIZEOF_INT], byteorder='little')
    return offset + file_length + SIZEOF_INT


@uu.report_to_backtrace
def HandleCrashHeader(data):
    dir_name, dir_name_end_idx = ReadStringFromData(data, 0)
    # log('Directory Name: {0}'.format(dir_name))

    file_name, file_name_end_idx = ReadStringFromData(data, dir_name_end_idx)
    # log('File Name: {0}'.format(file_name))

    return file_name_end_idx + (SIZEOF_INT * 3)


@uu.report_to_backtrace
def FindDumpFile(data, start_idx):
    current_idx = start_idx
    while current_idx < len(data):
        file_name, current_idx = ReadStringFromData(data, current_idx)
        # log('File Name Found: {0}'.format(file_name))

        if file_name.endswith('.dmp'):
            file_contents, current_idx = GetFileFromData(data, current_idx)
            return file_contents
        else:
            current_idx = SkipFileInData(data, current_idx)
        current_idx += 4
        # log('Current Index: {0}, Total Bytes: {1}'.format(
        # current_idx, len(data)))

    log('ERROR: No dump file found in data stream')
    return None


@uu.report_to_backtrace
def get_all_files(data, start_idx):
    r = {}

    current_idx = start_idx
    while current_idx < len(data):
        file_name, current_idx = ReadStringFromData(data, current_idx)
        if current_idx is None:
            break
        file_contents, current_idx = GetFileFromData(data, current_idx)
        r[file_name] = file_contents

        current_idx += 4

    return r

# ------------------------------------------------------------------------- #


class State:
    bytes_sent = 0
    bytes_received = 0
    requests_received = 0
    requests_processed = 0
    requests_dropped = 0

@uu.report_to_backtrace
def handleMultiPartPost(content_type, files):
    if content_type.startswith('multipart/form_data'):
        return None
    keys = files.keys()
    all_files = {}
    for key in keys:
        file = files[key][0]
        all_files[file['filename']] = file['body']
    dump_file_contents = b''
    try:
        fileKeys = all_files.keys()
        dumpKey = [f for f in fileKeys if f.lower().endswith('dmp')][0]
        dump_file_contents = all_files[dumpKey]
    except IndexError:
        return None, None
    return dump_file_contents, all_files


@uu.report_to_backtrace
def handleOctetStream(content_type, body, dump_last_request=None):
    if not content_type.startswith('application/octet-stream'):
        if not content_type.startswith('application/x-www-form-urlencoded'):
            return None, None
    _data = zlib.decompress(body)

    # Write the raw file to disk
    if dump_last_request:
        with open('%s/last_request.raw' % State.instance_dir, 'wb') as f:
            f.write(body)

    current_idx = HandleCrashHeader(_data)
    all_files = get_all_files(_data, current_idx)
    dump_file_contents = FindDumpFile(_data, current_idx)
    return dump_file_contents, all_files


class CrashReceiverHandler(tornado.web.RequestHandler):
    count = uu.timestamp_ms()

    def initialize(self, **data):
        self.config = data['config']
        self.uploader = data['uploader']

    @tornado.gen.coroutine
    @uu.report_to_backtrace
    def post(self, name, token):
        to_log = 'Request from {address}{fwd}. Size {size}'
        fwd = ''
        if 'X-Forwarded-For' in self.request.headers:
            fwd = ' (forwarded for: %s)' %\
                self.request.headers['X-Forwarded-For']

        log(to_log.format(address=self.request.remote_ip,
                          fwd=fwd, size=len(self.request.body)))

        State.bytes_received += len(self.request.body)
        State.requests_received += 1

        for el in self.config.blacklist:
            if re.search(el.regex, name):
                State.requests_dropped += 1
                log('Dropping request to %s (matches blacklist term /%s/)' % (name, el.regex))
                self.set_status(401)
                self.write('Unauthorized!')
                return

        # Check if we're at a very large queue size
        queue_length = self.uploader.get_queue_length()
        if queue_length >= self.config.limits.max_unsent_queue_size:
            log("Dropping the request, queue too large (%d)" % queue_length)
            State.requests_dropped += 1
            self.set_status(501)
            self.write("Server busy!")
            return

        ctype = self.request.headers.get('Content-Type', None)
        if ctype is None:
            self.set_status(400)
            self.write('Missing content-type header!')
            return

        dump_file_contents, all_files = handleMultiPartPost(
            ctype, self.request.files)
        if dump_file_contents is None:
            dump_file_contents, all_files = handleOctetStream(
                ctype, self.request.body, self.config.dump_last_request)
        if dump_file_contents is None:
            State.requests_dropped += 1
            self.set_status(404)
            self.write("Error!")
            return

        State.requests_processed += 1

        def get_url_format():
            if name in self.config.endpoint.custom_mapping.dict():
                return self.config.endpoint.custom_mapping.dict()[name]
            return self.config.endpoint.url_format

        url = get_url_format().format(**{
            'name': name,
            'suffix': self.get_argument('endpoint_suffix',
                                        self.config.endpoint.suffix),
            'port': self.get_argument('endpoint_port',
                                      self.config.endpoint.port),
            'token': token,
        })

        CrashReceiverHandler.count += 1
        num = CrashReceiverHandler.count

        flags = os.O_CREAT | os.O_EXCL | os.O_WRONLY

        def write_file(data, ext, mode):
            fd = os.open('%s/%s.%s' % (State.pending_dir, num, ext), flags)
            with os.fdopen(fd, mode) as file:
                file.write(data)

        extracted = {}

        # Pull in URL parameters
        for key in self.request.arguments:
            extracted[key] = self.request.arguments[key][0].decode('ascii')

        for xml_name in [x for x in all_files if x.endswith('xml')]:
            xmldata = all_files[xml_name].decode('utf-8')
            extracted.update(uxml.extract_attributes_from_xml(
                xmldata, self.config.attributes))

            # allow for multiple XML parsings
            # if len(extracted) > 0:
            #    break

        files = {}
        idx = 1
        for name in all_files:
            file_name = '%s_%s' % (num, idx)

            files[file_name] = name

            fd = os.open('%s/%s' % (State.pending_dir, file_name), flags)
            with os.fdopen(fd, 'wb') as file:
                file.write(all_files[name])

            idx += 1

        request_state = {
            'attributes': extracted,
            'files': files,
            'url': url
        }

        write_file(json.dumps(request_state), 'req', 'w')

        log('Saved dump %s' % num)

        self.uploader.enqueue(num)

        self.write('Thank you!')


class HealthHandler(tornado.web.RequestHandler):
    @tornado.gen.coroutine
    @uu.report_to_backtrace
    def get(self):
        self.set_status(200)
        self.write('OK')


class MetricsHandler(tornado.web.RequestHandler):
    def initialize(self, **data):
        self.config = data['config']
        self.uploader = data['uploader']

    @tornado.gen.coroutine
    @uu.report_to_backtrace
    def get(self):
        try:
            token = self.get_argument('token')
            if not token or len(token) < 1:
                self.set_status(400)
                self.write('Bad request!')
                return
        except BaseException:
            self.set_status(400)
            self.write('Bad request!')
            return

        if token != self.config.metrics.api_token:
            self.set_status(403)
            self.write('Forbidden!')
            return

        self.set_header('Content-Type', 'application/json')

        fields = {
            'bytes_sent': State.bytes_sent,
            'bytes_received': State.bytes_received,
            'requests_received': State.requests_received,
            'requests_processed': State.requests_processed,
            'requests_dropped': State.requests_dropped,
            'send_queue_size': self.uploader.get_queue_length(),
        }

        for k, v in fields.items():
            fields[k] = {
                '_type': 'i',
                '_value': v,
            }

        self.write(json.dumps(fields))

@uu.report_to_backtrace
def make_server(app, server_conf, socket, config):
    server = None

    if server_conf.ssl is None:
        server = tornado.httpserver.HTTPServer(
            app, max_buffer_size=config.max_request_size)
    else:
        server = tornado.httpserver.HTTPServer(
            app,
            max_buffer_size=config.max_request_size,
            ssl_options={
                "certfile": server_conf.ssl.certfile,
                "keyfile": server_conf.ssl.keyfile,
            })

    server.add_sockets(socket)

    return server


@uu.report_to_backtrace
def run_app(config):
    def bind_socket(conf, add_to_port=0):
        final_port = conf.listen_port+add_to_port
        log('binding {address}:{port}'.format(address=conf.address,
                                              port=final_port))
        return tornado.netutil.bind_sockets(final_port,
                                            address=conf.address,
                                            reuse_port=True,
                                            backlog=conf.listen_backlog)

    sockets = [(bind_socket(conf), conf) for conf in config.http]

    instance_id = tornado.process.fork_processes(config.processes)

    async def post_fork():
        pending_dir = '{dump_dirs}/{instance_id}/pending'.format(**config.dict(), instance_id=instance_id)
        sending_dir = '{dump_dirs}/{instance_id}/sending'.format(**config.dict(), instance_id=instance_id)

        State.pending_dir = pending_dir
        State.sending_dir = sending_dir
        State.instance_dir = '{dump_dirs}/{instance_id}/'.format(**config.dict(), instance_id=instance_id)

        uploader = Uploader(config, pending_dir, sending_dir)
        application = tornado.web.Application([
            (r'/post/(.*)/(.*)', CrashReceiverHandler, {'config': config, 'uploader': uploader}),
            (r'/health', HealthHandler),
            (r'/api/metrics', MetricsHandler, {'config': config, 'uploader': uploader}),
        ])

        servers = [make_server(application, server_conf, socket, config) for socket, server_conf in sockets]

        for file_name in glob('%s/*' % pending_dir):
            os.remove(file_name)

        try:
            os.makedirs(pending_dir)
        except FileExistsError:
            pass

        try:
            os.makedirs(sending_dir)
        except FileExistsError:
            pass

        def chown_dir(dir):
            uu.chown_dir(dir, config.run_as.username, config.run_as.group)

        chown_dir('%s/../..' % pending_dir)
        chown_dir('%s/..' % pending_dir)
        chown_dir(pending_dir)
        chown_dir(sending_dir)

        metrics_socket = bind_socket(config.metrics, instance_id)
        make_server(application, config.metrics, metrics_socket, config)

        uu.drop_privileges(config.run_as.username, config.run_as.group)

        log('starting the event loop... (instance %d)' % instance_id)
        for server in servers:
            server.start()

        uploader.start()

        await asyncio.Event().wait()

    loop = asyncio.get_event_loop()
    loop.run_until_complete(post_fork())

